#include <chrono>
#include <cstdlib>
#include <iostream>

/* OpenMP header */
#include <omp.h>

#include "is-prime.h"

namespace
{
   constexpr size_t padding_factor = 16;
   
   int64_t local_counter = 0;
   #pragma omp threadprivate(local_counter)
}

int main(int argc, char** argv)
{
   // read up to what max number we are going to count the prime numbers
   if(argc < 2) return EXIT_FAILURE;
   const int64_t limit = std::atoi(argv[1]);
   
   // how many OMP threads are we to create?
   int num_threads = 1;
   if(argc >= 3) num_threads = std::atoi(argv[2]);
   
   std::cout << "How many prime numbers p are there with p < "
             << limit << "? (" << num_threads << " OpenMP threads)\n\n";
   std::cout.flush();
   
   int64_t* counted_primes = new int64_t[padding_factor*num_threads];
   omp_set_num_threads(num_threads);  // optional; default: use $OMP_NUM_THREADS

   #pragma omp parallel
   {
      local_counter = 0;
      int thread_id = omp_get_thread_num();
      
      for(int64_t n = 6*(thread_id+1) - 1; n < limit; n += 6*num_threads)
         if(is_prime(n)) local_counter++;
      for(int64_t n = 6*(thread_id+1) + 1; n < limit; n += 6*num_threads)
         if(is_prime(n)) local_counter++;
      
      counted_primes[padding_factor*thread_id] = local_counter;
   }
      
   int64_t overall_primes = 0;
   if(limit >= 2) overall_primes++;  // 2 is a prime number
   if(limit >= 3) overall_primes++;  // 3 is a prime number
   
   for(int i = 0; i < num_threads; i++)
   {
      overall_primes += counted_primes[padding_factor*i];
      std::clog << "\tThread " << i << " counted\t" << counted_primes[padding_factor*i] << "\tprimes.\n";
   }

   std::cout << "\tThere are also\t2\tprimes below five.\n";
   
   std::cout << "\nTotal: There are " << overall_primes << " prime numbers smaller than " << limit << ".\n";
   
   delete[] counted_primes;
}
